@@ -26,49 +26,25 @@
 #include <stdlib.h>
 #include <string.h>
 
-#include <compiz.h>
+#include <compiz-core.h>
 
-#define FADE_SPEED_DEFAULT    5.0f
-#define FADE_SPEED_MIN        0.1f
-#define FADE_SPEED_MAX       10.0f
-#define FADE_SPEED_PRECISION  0.1f
-
-static char *winType[] = {
-    N_("Dock"),
-    N_("Toolbar"),
-    N_("Menu"),
-    N_("Utility"),
-    N_("Splash"),
-    N_("Normal"),
-    N_("Dialog"),
-    N_("ModalDialog"),
-    N_("DropdownMenu"),
-    N_("PopupMenu"),
-    N_("Tooltip"),
-    N_("Notification"),
-    N_("Combo"),
-    N_("Dnd"),
-    N_("Unknown")
-};
-#define N_WIN_TYPE (sizeof (winType) / sizeof (winType[0]))
-
-#define FADE_VISUAL_BELL_DEFAULT FALSE
-
-#define FADE_FULLSCREEN_VISUAL_BELL_DEFAULT FALSE
+static CompMetadata fadeMetadata;
 
 static int displayPrivateIndex;
 
 typedef struct _FadeDisplay {
-    int		    screenPrivateIndex;
-    HandleEventProc handleEvent;
-    int		    displayModals;
+    int			       screenPrivateIndex;
+    HandleEventProc	       handleEvent;
+    MatchExpHandlerChangedProc matchExpHandlerChanged;
+    int			       displayModals;
 } FadeDisplay;
 
 #define FADE_SCREEN_OPTION_FADE_SPEED		  0
-#define FADE_SCREEN_OPTION_WINDOW_TYPE		  1
+#define FADE_SCREEN_OPTION_WINDOW_MATCH		  1
 #define FADE_SCREEN_OPTION_VISUAL_BELL		  2
 #define FADE_SCREEN_OPTION_FULLSCREEN_VISUAL_BELL 3
-#define FADE_SCREEN_OPTION_NUM			  4
+#define FADE_SCREEN_OPTION_MINIMIZE_OPEN_CLOSE	  4
+#define FADE_SCREEN_OPTION_NUM			  5
 
 typedef struct _FadeScreen {
     int			   windowPrivateIndex;
@@ -82,7 +58,7 @@ typedef struct _FadeScreen {
     FocusWindowProc	   focusWindow;
     WindowResizeNotifyProc windowResizeNotify;
 
-    int wMask;
+    CompMatch match;
 } FadeScreen;
 
 typedef struct _FadeWindow {
@@ -96,24 +72,25 @@ typedef struct _FadeWindow {
     int unmapCnt;
 
     Bool shaded;
+    Bool fadeOut;
 
     int steps;
 } FadeWindow;
 
-#define GET_FADE_DISPLAY(d)				     \
-    ((FadeDisplay *) (d)->privates[displayPrivateIndex].ptr)
+#define GET_FADE_DISPLAY(d)					  \
+    ((FadeDisplay *) (d)->base.privates[displayPrivateIndex].ptr)
 
 #define FADE_DISPLAY(d)			   \
     FadeDisplay *fd = GET_FADE_DISPLAY (d)
 
-#define GET_FADE_SCREEN(s, fd)					 \
-    ((FadeScreen *) (s)->privates[(fd)->screenPrivateIndex].ptr)
+#define GET_FADE_SCREEN(s, fd)					      \
+    ((FadeScreen *) (s)->base.privates[(fd)->screenPrivateIndex].ptr)
 
 #define FADE_SCREEN(s)							\
     FadeScreen *fs = GET_FADE_SCREEN (s, GET_FADE_DISPLAY (s->display))
 
-#define GET_FADE_WINDOW(w, fs)				         \
-    ((FadeWindow *) (w)->privates[(fs)->windowPrivateIndex].ptr)
+#define GET_FADE_WINDOW(w, fs)					      \
+    ((FadeWindow *) (w)->base.privates[(fs)->windowPrivateIndex].ptr)
 
 #define FADE_WINDOW(w)					     \
     FadeWindow *fw = GET_FADE_WINDOW  (w,		     \
@@ -122,8 +99,21 @@ typedef struct _FadeWindow {
 
 #define NUM_OPTIONS(s) (sizeof ((s)->opt) / sizeof (CompOption))
 
+static void
+fadeUpdateWindowFadeMatch (CompDisplay     *display,
+			   CompOptionValue *value,
+			   CompMatch       *match)
+{
+    matchFini (match);
+    matchInit (match);
+    matchAddFromString (match, "!type=desktop");
+    matchAddGroup (match, MATCH_OP_AND_MASK, &value->match);
+    matchUpdate (display, match);
+}
+
 static CompOption *
-fadeGetScreenOptions (CompScreen *screen,
+fadeGetScreenOptions (CompPlugin *plugin,
+		      CompScreen *screen,
 		      int	 *count)
 {
     FADE_SCREEN (screen);
@@ -133,8 +123,9 @@ fadeGetScreenOptions (CompScreen *screen,
 }
 
 static Bool
-fadeSetScreenOption (CompScreen      *screen,
-		     char	     *name,
+fadeSetScreenOption (CompPlugin      *plugin,
+		     CompScreen      *screen,
+		     const char	     *name,
 		     CompOptionValue *value)
 {
     CompOption *o;
@@ -154,71 +145,22 @@ fadeSetScreenOption (CompScreen      *screen,
 	    return TRUE;
 	}
 	break;
-    case FADE_SCREEN_OPTION_WINDOW_TYPE:
-	if (compSetOptionList (o, value))
+    case FADE_SCREEN_OPTION_WINDOW_MATCH:
+	if (compSetMatchOption (o, value))
 	{
-	    fs->wMask = compWindowTypeMaskFromStringList (&o->value);
-	    fs->wMask &= ~CompWindowTypeDesktopMask;
+	    fadeUpdateWindowFadeMatch (screen->display, &o->value, &fs->match);
 	    return TRUE;
 	}
 	break;
-    case FADE_SCREEN_OPTION_VISUAL_BELL:
-    case FADE_SCREEN_OPTION_FULLSCREEN_VISUAL_BELL:
-	if (compSetBoolOption (o, value))
-	    return TRUE;
     default:
+	if (compSetOption (o, value))
+	    return TRUE;
 	break;
     }
 
     return FALSE;
 }
 
-static void
-fadeScreenInitOptions (FadeScreen *fs)
-{
-    CompOption *o;
-    int	       i;
-
-    o = &fs->opt[FADE_SCREEN_OPTION_FADE_SPEED];
-    o->name		= "fade_speed";
-    o->shortDesc	= N_("Fade Speed");
-    o->longDesc		= N_("Window fade speed");
-    o->type		= CompOptionTypeFloat;
-    o->value.f		= FADE_SPEED_DEFAULT;
-    o->rest.f.min	= FADE_SPEED_MIN;
-    o->rest.f.max	= FADE_SPEED_MAX;
-    o->rest.f.precision = FADE_SPEED_PRECISION;
-
-    o = &fs->opt[FADE_SCREEN_OPTION_WINDOW_TYPE];
-    o->name	         = "window_types";
-    o->shortDesc         = N_("Window Types");
-    o->longDesc	         = N_("Window types that should be fading");
-    o->type	         = CompOptionTypeList;
-    o->value.list.type   = CompOptionTypeString;
-    o->value.list.nValue = N_WIN_TYPE;
-    o->value.list.value  = malloc (sizeof (CompOptionValue) * N_WIN_TYPE);
-    for (i = 0; i < N_WIN_TYPE; i++)
-	o->value.list.value[i].s = strdup (winType[i]);
-    o->rest.s.string     = windowTypeString;
-    o->rest.s.nString    = nWindowTypeString;
-
-    fs->wMask = compWindowTypeMaskFromStringList (&o->value);
-
-    o = &fs->opt[FADE_SCREEN_OPTION_VISUAL_BELL];
-    o->name	  = "visual_bell";
-    o->shortDesc  = N_("Visual Bell");
-    o->longDesc	  = N_("Fade effect on system beep");
-    o->type	  = CompOptionTypeBool;
-    o->value.b    = FADE_VISUAL_BELL_DEFAULT;
-
-    o = &fs->opt[FADE_SCREEN_OPTION_FULLSCREEN_VISUAL_BELL];
-    o->name	  = "fullscreen_visual_bell";
-    o->shortDesc  = N_("Fullscreen Visual Bell");
-    o->longDesc	  = N_("Fullscreen fade effect on system beep");
-    o->type	  = CompOptionTypeBool;
-    o->value.b    = FADE_FULLSCREEN_VISUAL_BELL_DEFAULT;
-}
-
 static void
 fadePreparePaintScreen (CompScreen *s,
 			int	   msSinceLastPaint)
@@ -261,6 +203,7 @@ fadeWindowStop (CompWindow *w)
 static Bool
 fadePaintWindow (CompWindow		 *w,
 		 const WindowPaintAttrib *attrib,
+		 const CompTransform	 *transform,
 		 Region			 region,
 		 unsigned int		 mask)
 {
@@ -273,12 +216,17 @@ fadePaintWindow (CompWindow		 *w,
     if (!w->screen->canDoSlightlySaturated)
 	fw->saturation = attrib->saturation;
 
-    if (fw->opacity    != attrib->opacity    ||
+    if (fw->destroyCnt			     ||
+	fw->unmapCnt			     ||
+	fw->opacity    != attrib->opacity    ||
 	fw->brightness != attrib->brightness ||
 	fw->saturation != attrib->saturation)
     {
 	WindowPaintAttrib fAttrib = *attrib;
 
+	if (fw->fadeOut)
+	    fAttrib.opacity = 0;
+
 	if (fw->steps)
 	{
 	    GLint opacity;
@@ -286,21 +234,21 @@ fadePaintWindow (CompWindow		 *w,
 	    GLint saturation;
 
 	    opacity = fw->opacity;
-	    if (attrib->opacity > fw->opacity)
+	    if (fAttrib.opacity > fw->opacity)
 	    {
 		opacity = fw->opacity + fw->steps;
-		if (opacity > attrib->opacity)
-		    opacity = attrib->opacity;
+		if (opacity > fAttrib.opacity)
+		    opacity = fAttrib.opacity;
 	    }
-	    else if (attrib->opacity < fw->opacity)
+	    else if (fAttrib.opacity < fw->opacity)
 	    {
 		if (w->type & CompWindowTypeUnknownMask)
 		    opacity = fw->opacity - (fw->steps >> 1);
 		else
 		    opacity = fw->opacity - fw->steps;
 
-		if (opacity < attrib->opacity)
-		    opacity = attrib->opacity;
+		if (opacity < fAttrib.opacity)
+		    opacity = fAttrib.opacity;
 	    }
 
 	    brightness = fw->brightness;
@@ -357,13 +305,13 @@ fadePaintWindow (CompWindow		 *w,
 	fAttrib.saturation = fw->saturation;
 
 	UNWRAP (fs, s, paintWindow);
-	status = (*s->paintWindow) (w, &fAttrib, region, mask);
+	status = (*s->paintWindow) (w, &fAttrib, transform, region, mask);
 	WRAP (fs, s, paintWindow, fadePaintWindow);
     }
     else
     {
 	UNWRAP (fs, s, paintWindow);
-	status = (*s->paintWindow) (w, attrib, region, mask);
+	status = (*s->paintWindow) (w, attrib, transform, region, mask);
 	WRAP (fs, s, paintWindow, fadePaintWindow);
     }
 
@@ -461,18 +409,21 @@ fadeHandleEvent (CompDisplay *d,
 	{
 	    FADE_SCREEN (w->screen);
 
-	    if (w->texture->pixmap && (fs->wMask & w->type))
+	    if (!fs->opt[FADE_SCREEN_OPTION_MINIMIZE_OPEN_CLOSE].value.b)
+		break;
+
+	    if (w->texture->pixmap && matchEval (&fs->match, w))
 	    {
 		FADE_WINDOW (w);
 
-		w->paint.opacity = 0;
-
 		if (fw->opacity == 0xffff)
 		    fw->opacity = 0xfffe;
 
 		fw->destroyCnt++;
 		w->destroyRefCnt++;
 
+		fw->fadeOut = TRUE;
+
 		addWindowDamage (w);
 	    }
 
@@ -488,16 +439,19 @@ fadeHandleEvent (CompDisplay *d,
 
 	    fw->shaded = w->shaded;
 
-	    if (!fw->shaded && w->texture->pixmap && (fs->wMask & w->type))
-	    {
-		w->paint.opacity = 0;
+	    if (!fs->opt[FADE_SCREEN_OPTION_MINIMIZE_OPEN_CLOSE].value.b)
+		break;
 
+	    if (!fw->shaded && w->texture->pixmap && matchEval (&fs->match, w))
+	    {
 		if (fw->opacity == 0xffff)
 		    fw->opacity = 0xfffe;
 
 		fw->unmapCnt++;
 		w->unmapRefCnt++;
 
+		fw->fadeOut = TRUE;
+
 		addWindowDamage (w);
 	    }
 
@@ -508,10 +462,10 @@ fadeHandleEvent (CompDisplay *d,
 	w = findWindowAtDisplay (d, event->xmap.window);
 	if (w)
 	{
-	    if (!(w->type & CompWindowTypeDesktopMask))
-		w->paint.opacity = getWindowProp32 (d, w->id,
-						    d->winOpacityAtom,
-						    OPAQUE);
+	    FADE_SCREEN(w->screen);
+
+	    if (!fs->opt[FADE_SCREEN_OPTION_MINIMIZE_OPEN_CLOSE].value.b)
+		break;
 
 	    fadeWindowStop (w);
 
@@ -613,13 +567,18 @@ fadeDamageWindowRect (CompWindow *w,
     {
 	FADE_WINDOW (w);
 
+	fw->fadeOut = FALSE;
+
 	if (fw->shaded)
 	{
 	    fw->shaded = w->shaded;
 	}
-	else if ((fs->wMask & w->type) && fw->opacity == w->paint.opacity)
+	else if (matchEval (&fs->match, w))
 	{
-	    fw->opacity = 0;
+	    if (fs->opt[FADE_SCREEN_OPTION_MINIMIZE_OPEN_CLOSE].value.b)
+	    {
+		fw->opacity = 0;
+	    }
 	}
     }
 
@@ -649,7 +608,11 @@ fadeFocusWindow (CompWindow *w)
 }
 
 static void
-fadeWindowResizeNotify (CompWindow *w)
+fadeWindowResizeNotify (CompWindow *w,
+			int	   dx,
+			int	   dy,
+			int	   dwidth,
+			int	   dheight)
 {
     FADE_SCREEN (w->screen);
 
@@ -657,16 +620,34 @@ fadeWindowResizeNotify (CompWindow *w)
 	fadeWindowStop (w);
 
     UNWRAP (fs, w->screen, windowResizeNotify);
-    (*w->screen->windowResizeNotify) (w);
+    (*w->screen->windowResizeNotify) (w, dx, dy, dwidth, dheight);
     WRAP (fs, w->screen, windowResizeNotify, fadeWindowResizeNotify);
 }
 
+static void
+fadeMatchExpHandlerChanged (CompDisplay *d)
+{
+    CompScreen *s;
+
+    FADE_DISPLAY (d);
+
+    for (s = d->screens; s; s = s->next)
+	matchUpdate (d, &GET_FADE_SCREEN (s,fd)->match);
+
+    UNWRAP (fd, d, matchExpHandlerChanged);
+    (*d->matchExpHandlerChanged) (d);
+    WRAP (fd, d, matchExpHandlerChanged, fadeMatchExpHandlerChanged);
+}
+
 static Bool
 fadeInitDisplay (CompPlugin  *p,
 		 CompDisplay *d)
 {
     FadeDisplay *fd;
 
+    if (!checkPluginABI ("core", CORE_ABIVERSION))
+	return FALSE;
+
     fd = malloc (sizeof (FadeDisplay));
     if (!fd)
 	return FALSE;
@@ -681,14 +662,15 @@ fadeInitDisplay (CompPlugin  *p,
     fd->displayModals = 0;
 
     WRAP (fd, d, handleEvent, fadeHandleEvent);
+    WRAP (fd, d, matchExpHandlerChanged, fadeMatchExpHandlerChanged);
 
-    d->privates[displayPrivateIndex].ptr = fd;
+    d->base.privates[displayPrivateIndex].ptr = fd;
 
     return TRUE;
 }
 
 static void
-fadeFiniDisplay (CompPlugin *p,
+fadeFiniDisplay (CompPlugin  *p,
 		 CompDisplay *d)
 {
     FADE_DISPLAY (d);
@@ -696,10 +678,19 @@ fadeFiniDisplay (CompPlugin *p,
     freeScreenPrivateIndex (d, fd->screenPrivateIndex);
 
     UNWRAP (fd, d, handleEvent);
+    UNWRAP (fd, d, matchExpHandlerChanged);
 
     free (fd);
 }
 
+static const CompMetadataOptionInfo fadeScreenOptionInfo[] = {
+    { "fade_speed", "float", "<min>0.1</min>", 0, 0 },
+    { "window_match", "match", "<helper>true</helper>", 0, 0 },
+    { "visual_bell", "bool", 0, 0, 0 },
+    { "fullscreen_visual_bell", "bool", 0, 0, 0 },
+    { "minimize_open_close", "bool", 0, 0, 0 }
+};
+
 static Bool
 fadeInitScreen (CompPlugin *p,
 		CompScreen *s)
@@ -712,16 +703,31 @@ fadeInitScreen (CompPlugin *p,
     if (!fs)
 	return FALSE;
 
-    fs->windowPrivateIndex = allocateWindowPrivateIndex (s);
-    if (fs->windowPrivateIndex < 0)
+    if (!compInitScreenOptionsFromMetadata (s,
+					    &fadeMetadata,
+					    fadeScreenOptionInfo,
+					    fs->opt,
+					    FADE_SCREEN_OPTION_NUM))
     {
 	free (fs);
 	return FALSE;
     }
 
-    fs->fadeTime = 1000.0f / FADE_SPEED_DEFAULT;
+    fs->windowPrivateIndex = allocateWindowPrivateIndex (s);
+    if (fs->windowPrivateIndex < 0)
+    {
+	compFiniScreenOptions (s, fs->opt, FADE_SCREEN_OPTION_NUM);
+	free (fs);
+	return FALSE;
+    }
 
-    fadeScreenInitOptions (fs);
+    fs->fadeTime = 1000.0f / fs->opt[FADE_SCREEN_OPTION_FADE_SPEED].value.f;
+
+    matchInit (&fs->match);
+
+    fadeUpdateWindowFadeMatch (s->display,
+			       &fs->opt[FADE_SCREEN_OPTION_WINDOW_MATCH].value,
+			       &fs->match);
 
     WRAP (fs, s, preparePaintScreen, fadePreparePaintScreen);
     WRAP (fs, s, paintWindow, fadePaintWindow);
@@ -729,7 +735,7 @@ fadeInitScreen (CompPlugin *p,
     WRAP (fs, s, focusWindow, fadeFocusWindow);
     WRAP (fs, s, windowResizeNotify, fadeWindowResizeNotify);
 
-    s->privates[fd->screenPrivateIndex].ptr = fs;
+    s->base.privates[fd->screenPrivateIndex].ptr = fs;
 
     return TRUE;
 }
@@ -740,6 +746,8 @@ fadeFiniScreen (CompPlugin *p,
 {
     FADE_SCREEN (s);
 
+    matchFini (&fs->match);
+
     freeWindowPrivateIndex (s, fs->windowPrivateIndex);
 
     UNWRAP (fs, s, preparePaintScreen);
@@ -748,6 +756,8 @@ fadeFiniScreen (CompPlugin *p,
     UNWRAP (fs, s, focusWindow);
     UNWRAP (fs, s, windowResizeNotify);
 
+    compFiniScreenOptions (s, fs->opt, FADE_SCREEN_OPTION_NUM);
+
     free (fs);
 }
 
@@ -772,8 +782,9 @@ fadeInitWindow (CompPlugin *p,
     fw->destroyCnt = 0;
     fw->unmapCnt   = 0;
     fw->shaded     = w->shaded;
+    fw->fadeOut    = FALSE;
 
-    w->privates[fs->windowPrivateIndex].ptr = fw;
+    w->base.privates[fs->windowPrivateIndex].ptr = fw;
 
     if (w->attrib.map_state == IsViewable)
     {
@@ -793,15 +804,88 @@ fadeFiniWindow (CompPlugin *p,
     fadeRemoveDisplayModal (w->screen->display, w);
     fadeWindowStop (w);
 
+    w->paint.opacity    = w->opacity;
+    w->paint.brightness = w->brightness;
+    w->paint.saturation = w->saturation;
+
     free (fw);
 }
 
+static CompBool
+fadeInitObject (CompPlugin *p,
+		CompObject *o)
+{
+    static InitPluginObjectProc dispTab[] = {
+	(InitPluginObjectProc) 0, /* InitCore */
+	(InitPluginObjectProc) fadeInitDisplay,
+	(InitPluginObjectProc) fadeInitScreen,
+	(InitPluginObjectProc) fadeInitWindow
+    };
+
+    RETURN_DISPATCH (o, dispTab, ARRAY_SIZE (dispTab), TRUE, (p, o));
+}
+
+static void
+fadeFiniObject (CompPlugin *p,
+		CompObject *o)
+{
+    static FiniPluginObjectProc dispTab[] = {
+	(FiniPluginObjectProc) 0, /* FiniCore */
+	(FiniPluginObjectProc) fadeFiniDisplay,
+	(FiniPluginObjectProc) fadeFiniScreen,
+	(FiniPluginObjectProc) fadeFiniWindow
+    };
+
+    DISPATCH (o, dispTab, ARRAY_SIZE (dispTab), (p, o));
+}
+
+static CompOption *
+fadeGetObjectOptions (CompPlugin *plugin,
+		      CompObject *object,
+		      int	 *count)
+{
+    static GetPluginObjectOptionsProc dispTab[] = {
+	(GetPluginObjectOptionsProc) 0, /* GetCoreOptions */
+	(GetPluginObjectOptionsProc) 0, /* GetDisplayOptions */
+	(GetPluginObjectOptionsProc) fadeGetScreenOptions
+    };
+
+    RETURN_DISPATCH (object, dispTab, ARRAY_SIZE (dispTab),
+		     (void *) (*count = 0), (plugin, object, count));
+}
+
+static CompBool
+fadeSetObjectOption (CompPlugin      *plugin,
+		     CompObject      *object,
+		     const char      *name,
+		     CompOptionValue *value)
+{
+    static SetPluginObjectOptionProc dispTab[] = {
+	(SetPluginObjectOptionProc) 0, /* SetCoreOption */
+	(SetPluginObjectOptionProc) 0, /* SetDisplayOption */
+	(SetPluginObjectOptionProc) fadeSetScreenOption
+    };
+
+    RETURN_DISPATCH (object, dispTab, ARRAY_SIZE (dispTab), FALSE,
+		     (plugin, object, name, value));
+}
+
 static Bool
 fadeInit (CompPlugin *p)
 {
+    if (!compInitPluginMetadataFromInfo (&fadeMetadata, p->vTable->name, 0, 0,
+					 fadeScreenOptionInfo,
+					 FADE_SCREEN_OPTION_NUM))
+	return FALSE;
+
     displayPrivateIndex = allocateDisplayPrivateIndex ();
     if (displayPrivateIndex < 0)
+    {
+	compFiniMetadata (&fadeMetadata);
 	return FALSE;
+    }
+
+    compAddMetadataFromFile (&fadeMetadata, p->vTable->name);
 
     return TRUE;
 }
@@ -809,47 +893,29 @@ fadeInit (CompPlugin *p)
 static void
 fadeFini (CompPlugin *p)
 {
-    if (displayPrivateIndex >= 0)
-	freeDisplayPrivateIndex (displayPrivateIndex);
+    freeDisplayPrivateIndex (displayPrivateIndex);
+    compFiniMetadata (&fadeMetadata);
 }
 
-static int
-fadeGetVersion (CompPlugin *plugin,
-		int	   version)
+static CompMetadata *
+fadeGetMetadata (CompPlugin *plugin)
 {
-    return ABIVERSION;
+    return &fadeMetadata;
 }
 
-CompPluginDep fadeDeps[] = {
-    { CompPluginRuleBefore, "cube" },
-    { CompPluginRuleBefore, "scale" }
-};
-
 static CompPluginVTable fadeVTable = {
     "fade",
-    N_("Fading Windows"),
-    N_("Fade in windows when mapped and fade out windows when unmapped"),
-    fadeGetVersion,
+    fadeGetMetadata,
     fadeInit,
     fadeFini,
-    fadeInitDisplay,
-    fadeFiniDisplay,
-    fadeInitScreen,
-    fadeFiniScreen,
-    fadeInitWindow,
-    fadeFiniWindow,
-    0, /* GetDisplayOptions */
-    0, /* SetDisplayOption */
-    fadeGetScreenOptions,
-    fadeSetScreenOption,
-    fadeDeps,
-    sizeof (fadeDeps) / sizeof (fadeDeps[0]),
-    0, /* Features */
-    0  /* nFeatures */
+    fadeInitObject,
+    fadeFiniObject,
+    fadeGetObjectOptions,
+    fadeSetObjectOption
 };
 
 CompPluginVTable *
-getCompPluginInfo (void)
+getCompPluginInfo20070830 (void)
 {
     return &fadeVTable;
 }
