@@ -24,30 +24,18 @@
 #include <stdlib.h>
 #include <string.h>
 
-#include <compiz.h>
+#include <compiz-core.h>
 
 #include <glib.h>
 
-#define PLACE_WORKAROUND_DEFAULT TRUE
+static CompMetadata placeMetadata;
 
-typedef enum {
-    PlaceModeCascade  = 0,
-    PlaceModeCentered = 1,
-    PlaceModeSmart    = 2,
-    PlaceModeMaximize = 3,
-    PlaceModeRandom   = 4
-} PlaceMode;
-
-static char *modeString[] = {
-    N_("Cascade"),
-    N_("Centered"),
-    N_("Smart"),
-    N_("Maximize"),
-    N_("Random")
-};
-static int nModeString = sizeof (modeString) / sizeof (modeString[0]);
-
-#define PLACE_MODE_DEFAULT (modeString[0])
+#define PLACE_MODE_CASCADE  0
+#define PLACE_MODE_CENTERED 1
+#define PLACE_MODE_SMART    2
+#define PLACE_MODE_MAXIMIZE 3
+#define PLACE_MODE_RANDOM   4
+#define PLACE_MODE_LAST     PLACE_MODE_RANDOM
 
 /* overlap types */
 #define NONE    0
@@ -58,7 +46,6 @@ static int displayPrivateIndex;
 
 typedef struct _PlaceDisplay {
     int		    screenPrivateIndex;
-    HandleEventProc handleEvent;
 } PlaceDisplay;
 
 #define PLACE_SCREEN_OPTION_WORKAROUND        0
@@ -74,19 +61,18 @@ typedef struct _PlaceDisplay {
 typedef struct _PlaceScreen {
     CompOption opt[PLACE_SCREEN_OPTION_NUM];
 
-    DamageWindowRectProc damageWindowRect;
-
-    PlaceMode placeMode;
+    PlaceWindowProc                 placeWindow;
+    ValidateWindowResizeRequestProc validateWindowResizeRequest;
 } PlaceScreen;
 
-#define GET_PLACE_DISPLAY(d)				      \
-    ((PlaceDisplay *) (d)->privates[displayPrivateIndex].ptr)
+#define GET_PLACE_DISPLAY(d)					   \
+    ((PlaceDisplay *) (d)->base.privates[displayPrivateIndex].ptr)
 
 #define PLACE_DISPLAY(d)		     \
     PlaceDisplay *pd = GET_PLACE_DISPLAY (d)
 
-#define GET_PLACE_SCREEN(s, pd)					  \
-    ((PlaceScreen *) (s)->privates[(pd)->screenPrivateIndex].ptr)
+#define GET_PLACE_SCREEN(s, pd)					       \
+    ((PlaceScreen *) (s)->base.privates[(pd)->screenPrivateIndex].ptr)
 
 #define PLACE_SCREEN(s)							   \
     PlaceScreen *ps = GET_PLACE_SCREEN (s, GET_PLACE_DISPLAY (s->display))
@@ -153,28 +139,9 @@ placeMatchViewport (CompWindow *w,
 			      y);
 }
 
-static void
-placeUpdateMode (CompScreen *screen)
-{
-    char *mode;
-    int  i;
-
-    PLACE_SCREEN (screen);
-
-    mode = ps->opt[PLACE_SCREEN_OPTION_MODE].value.s;
-
-    for (i = 0; i < sizeof (modeString) / sizeof (modeString[0]); i++)
-    {
-	if (strcmp (modeString[i], mode) == 0)
-	{
-	    ps->placeMode = i;
-	    break;
-	}
-    }
-}
-
 static CompOption *
-placeGetScreenOptions (CompScreen *screen,
+placeGetScreenOptions (CompPlugin *plugin,
+		       CompScreen *screen,
 		       int	  *count)
 {
     PLACE_SCREEN (screen);
@@ -184,8 +151,9 @@ placeGetScreenOptions (CompScreen *screen,
 }
 
 static Bool
-placeSetScreenOption (CompScreen      *screen,
-		      char	      *name,
+placeSetScreenOption (CompPlugin      *plugin,
+		      CompScreen      *screen,
+		      const char      *name,
 		      CompOptionValue *value)
 {
     CompOption *o;
@@ -198,16 +166,9 @@ placeSetScreenOption (CompScreen      *screen,
 	return FALSE;
 
     switch (index) {
-    case PLACE_SCREEN_OPTION_WORKAROUND:
-	if (compSetBoolOption (o, value))
-	    return TRUE;
-	break;
     case PLACE_SCREEN_OPTION_MODE:
-	if (compSetStringOption (o, value))
-	{
-	    placeUpdateMode (screen);
+	if (compSetIntOption (o, value))
 	    return TRUE;
-	}
 	break;
     case PLACE_SCREEN_OPTION_POSITION_MATCHES:
     case PLACE_SCREEN_OPTION_VIEWPORT_MATCHES:
@@ -219,110 +180,17 @@ placeSetScreenOption (CompScreen      *screen,
 		matchUpdate (screen->display, &o->value.list.value[i].match);
 
 	    return TRUE;
-	}
-	break;
-    case PLACE_SCREEN_OPTION_POSITION_X_VALUES:
-    case PLACE_SCREEN_OPTION_POSITION_Y_VALUES:
-    case PLACE_SCREEN_OPTION_VIEWPORT_X_VALUES:
-    case PLACE_SCREEN_OPTION_VIEWPORT_Y_VALUES:
-	if (compSetOptionList (o, value))
-	    return TRUE;
+       }
+       break;
     default:
+	if (compSetOption (o, value))
+	    return TRUE;
 	break;
     }
 
     return FALSE;
 }
 
-static void
-placeScreenInitOptions (PlaceScreen *ps)
-{
-    CompOption *o;
-
-    o = &ps->opt[PLACE_SCREEN_OPTION_WORKAROUND];
-    o->name	 = "workarounds";
-    o->shortDesc = N_("Workarounds");
-    o->longDesc	 = N_("Window placement workarounds");
-    o->type	 = CompOptionTypeBool;
-    o->value.b	 = PLACE_WORKAROUND_DEFAULT;
-
-    o = &ps->opt[PLACE_SCREEN_OPTION_MODE];
-    o->name	      = "mode";
-    o->shortDesc      = N_("Placement Mode");
-    o->longDesc	      = N_("Algorithm to use for window placement");
-    o->type	      = CompOptionTypeString;
-    o->value.s	      = strdup (PLACE_MODE_DEFAULT);
-    o->rest.s.string  = modeString;
-    o->rest.s.nString = nModeString;
-
-    o = &ps->opt[PLACE_SCREEN_OPTION_POSITION_MATCHES];
-    o->name	         = "position_matches";
-    o->shortDesc         = N_("Positioned windows");
-    o->longDesc	         = N_("Windows that should be positioned by default");
-    o->type	         = CompOptionTypeList;
-    o->value.list.type   = CompOptionTypeMatch;
-    o->value.list.nValue = 0;
-    o->value.list.value  = NULL;
-    o->rest.s.string     = NULL;
-    o->rest.s.nString    = 0;
-
-    o = &ps->opt[PLACE_SCREEN_OPTION_POSITION_X_VALUES];
-    o->name	         = "position_x_values";
-    o->shortDesc         = N_("X Positions");
-    o->longDesc	         = N_("X position values");
-    o->type	         = CompOptionTypeList;
-    o->value.list.type   = CompOptionTypeInt;
-    o->value.list.nValue = 0;
-    o->value.list.value  = NULL;
-    o->rest.i.min	 = MINSHORT;
-    o->rest.i.max	 = MAXSHORT;
-
-    o = &ps->opt[PLACE_SCREEN_OPTION_POSITION_Y_VALUES];
-    o->name	         = "position_y_values";
-    o->shortDesc         = N_("Y Positions");
-    o->longDesc	         = N_("Y position values");
-    o->type	         = CompOptionTypeList;
-    o->value.list.type   = CompOptionTypeInt;
-    o->value.list.nValue = 0;
-    o->value.list.value  = NULL;
-    o->rest.i.min	 = MINSHORT;
-    o->rest.i.max	 = MAXSHORT;
-
-    o = &ps->opt[PLACE_SCREEN_OPTION_VIEWPORT_MATCHES];
-    o->name	         = "viewport_matches";
-    o->shortDesc         = N_("Viewport positioned windows");
-    o->longDesc	         = N_("Windows that should positioned in specific "
-			      "viewports by default");
-    o->type	         = CompOptionTypeList;
-    o->value.list.type   = CompOptionTypeMatch;
-    o->value.list.nValue = 0;
-    o->value.list.value  = NULL;
-    o->rest.s.string     = NULL;
-    o->rest.s.nString    = 0;
-
-    o = &ps->opt[PLACE_SCREEN_OPTION_VIEWPORT_X_VALUES];
-    o->name	         = "viewport_x_values";
-    o->shortDesc         = N_("X Viewport Positions");
-    o->longDesc	         = N_("Horizontal viewport positions");
-    o->type	         = CompOptionTypeList;
-    o->value.list.type   = CompOptionTypeInt;
-    o->value.list.nValue = 0;
-    o->value.list.value  = NULL;
-    o->rest.i.min	 = 0;
-    o->rest.i.max	 = 32;
-
-    o = &ps->opt[PLACE_SCREEN_OPTION_VIEWPORT_Y_VALUES];
-    o->name	         = "viewport_y_values";
-    o->shortDesc         = N_("Y Viewport Positions");
-    o->longDesc	         = N_("Vertical viewport positions");
-    o->type	         = CompOptionTypeList;
-    o->value.list.type   = CompOptionTypeInt;
-    o->value.list.nValue = 0;
-    o->value.list.value  = NULL;
-    o->rest.i.min	 = 0;
-    o->rest.i.max	 = 32;
-}
-
 typedef enum {
     PlaceLeft,
     PlaceRight,
@@ -415,6 +283,16 @@ get_window_height (CompWindow *window)
     return window->serverHeight + window->serverBorderWidth * 2;
 }
 
+static void
+get_outer_rect_of_window (CompWindow *w,
+			  XRectangle *r)
+{
+    r->x      = w->serverX - w->input.left;
+    r->y      = w->serverY - w->input.top;
+    r->width  = get_window_width (w)  + w->input.left + w->input.right;
+    r->height = get_window_height (w) + w->input.top  + w->input.bottom;
+}
+
 static void
 find_next_cascade (CompWindow *window,
 		   GList      *windows,
@@ -563,8 +441,8 @@ find_most_freespace (CompWindow *window,
 
     get_workarea_of_current_output_device (window->screen, &work_area);
 
-    getOuterRectOfWindow (focus_window, &avoid);
-    getOuterRectOfWindow (window, &outer);
+    get_outer_rect_of_window (focus_window, &avoid);
+    get_outer_rect_of_window (window, &outer);
 
     /* Find the areas of choosing the various sides of the focus window */
     max_width  = MIN (avoid.width, outer.width);
@@ -705,7 +583,7 @@ rectangle_overlaps_some_window (XRectangle *rect,
 	case CompWindowTypeUtilMask:
 	case CompWindowTypeToolbarMask:
 	case CompWindowTypeMenuMask:
-	    getOuterRectOfWindow (other, &other_rect);
+	    get_outer_rect_of_window (other, &other_rect);
 
 	    if (rectangleIntersect (rect, &other_rect, &dest))
 		return TRUE;
@@ -826,7 +704,7 @@ find_first_fit (CompWindow *window,
     right_sorted = g_list_sort (right_sorted, topmost_cmp);
     right_sorted = g_list_sort (right_sorted, leftmost_cmp);
 
-    getOuterRectOfWindow (window, &rect);
+    get_outer_rect_of_window (window, &rect);
 
     get_workarea_of_current_output_device (window->screen, &work_area);
 
@@ -855,7 +733,7 @@ find_first_fit (CompWindow *window,
 	CompWindow *w = tmp->data;
 	XRectangle outer_rect;
 
-	getOuterRectOfWindow (w, &outer_rect);
+	get_outer_rect_of_window (w, &outer_rect);
 
 	rect.x = outer_rect.x;
 	rect.y = outer_rect.y + outer_rect.height;
@@ -881,7 +759,7 @@ find_first_fit (CompWindow *window,
 	CompWindow *w = tmp->data;
 	XRectangle outer_rect;
 
-	getOuterRectOfWindow (w, &outer_rect);
+	get_outer_rect_of_window (w, &outer_rect);
 
 	rect.x = outer_rect.x + outer_rect.width;
 	rect.y = outer_rect.y;
@@ -923,9 +801,18 @@ placeRandom (CompWindow *window,
 	     int	*x,
 	     int	*y)
 {
-    *x = workarea->x + rand () % (workarea->width - get_window_width (window));
-    *y = workarea->y + rand () % (workarea->height -
-				  get_window_height (window));
+    int remainX, remainY;
+
+    *x = workarea->x;
+    *y = workarea->y;
+
+    remainX = workarea->width - get_window_width (window);
+    if (remainX > 0)
+	*x += rand () % remainX;
+
+    remainY = workarea->height - get_window_height (window);
+    if (remainY > 0)
+	*y += rand () % remainY;
 }
 
 static void
@@ -964,8 +851,10 @@ placeSmart (CompWindow *window,
     xOptimal = xTmp; yOptimal = yTmp;
 
     /* client gabarit */
-    int ch = get_window_height (window) - 1;
-    int cw = get_window_width (window) - 1;
+    int ch = get_window_height (window) + window->input.top +
+	     window->input.bottom - 1;
+    int cw = get_window_width (window) + window->input.left +
+	     window->input.right - 1;
 
     /* loop over possible positions */
     do
@@ -993,12 +882,12 @@ placeSmart (CompWindow *window,
 				    CompWindowTypeDesktopMask)))
 		{
 
-		    xl = wi->attrib.x;
-		    yt = wi->attrib.y;
+		    xl = wi->attrib.x - wi->input.left;
+		    yt = wi->attrib.y - wi->input.top;
 		    xr = xl + get_window_width (wi) + window->input.left
-			+ window->input.right;
+			+ wi->input.right;
 		    yb = yt + get_window_height (wi) + window->input.top
-			+ window->input.bottom;
+			+ wi->input.bottom;
 
 		    /* if windows overlap, calc the overall overlapping */
 		    if ((cxl < xr) && (cxr > xl) &&
@@ -1041,7 +930,7 @@ placeSmart (CompWindow *window,
 	/* really need to loop? test if there's any overlap */
 	if (overlap > NONE)
 	{
-	    possible = workarea->width;
+	    possible = workarea->x + workarea->width;
 
 	    if (possible - cw > xTmp) possible -= cw;
 
@@ -1055,12 +944,12 @@ placeSmart (CompWindow *window,
 				    CompWindowTypeDesktopMask)))
 		{
 
-		    xl = wi->attrib.x;
-		    yt = wi->attrib.y;
-		    xr = xl + get_window_width (wi) + window->input.left
-			+ window->input.right;
-		    yb = yt + get_window_height (wi) + window->input.top
-			+ window->input.bottom;
+		    xl = wi->attrib.x - wi->input.left;
+		    yt = wi->attrib.y - wi->input.top;
+		    xr = xl + get_window_width (wi) + wi->input.left
+			+ wi->input.right;
+		    yb = yt + get_window_height (wi) + wi->input.top
+			+ wi->input.bottom;
 
 		    /* if not enough room above or under the current
 		     * client determine the first non-overlapped x position
@@ -1094,12 +983,12 @@ placeSmart (CompWindow *window,
 		    !(wi->wmType & (CompWindowTypeDockMask |
 				    CompWindowTypeDesktopMask)))
 		{
-		    xl = wi->attrib.x;
-		    yt = wi->attrib.y;
-		    xr = xl + get_window_width (wi) + window->input.left
-			+ window->input.right;
-		    yb = yt + get_window_height (wi) + window->input.top
-			+ window->input.bottom;
+		    xl = wi->attrib.x - wi->input.left;
+		    yt = wi->attrib.y - wi->input.top;
+		    xr = xl + get_window_width (wi) + wi->input.left
+			+ wi->input.right;
+		    yb = yt + get_window_height (wi) + wi->input.top
+			+ wi->input.bottom;
 
 		    /* if not enough room to the left or right of the current
 		     * client determine the first non-overlapped y position
@@ -1121,16 +1010,40 @@ placeSmart (CompWindow *window,
     if (ch >= workarea->height)
 	yOptimal = workarea->y;
 
-    *x = xOptimal;
-    *y = yOptimal;
+    *x = xOptimal + window->input.left;
+    *y = yOptimal + window->input.top;
 }
 
 static void
-placeWindow (CompWindow *window,
-	     int        x,
-	     int        y,
-	     int        *new_x,
-	     int        *new_y)
+placeSendWindowMaximizationRequest (CompWindow *w)
+{
+    XEvent      xev;
+    CompDisplay *d = w->screen->display;
+
+    xev.xclient.type    = ClientMessage;
+    xev.xclient.display = d->display;
+    xev.xclient.format  = 32;
+
+    xev.xclient.message_type = d->winStateAtom;
+    xev.xclient.window	     = w->id;
+
+    xev.xclient.data.l[0] = 1;
+    xev.xclient.data.l[1] = d->winStateMaximizedHorzAtom;
+    xev.xclient.data.l[2] = d->winStateMaximizedVertAtom;
+    xev.xclient.data.l[3] = 0;
+    xev.xclient.data.l[4] = 0;
+
+    XSendEvent (d->display, w->screen->root, FALSE,
+		SubstructureRedirectMask | SubstructureNotifyMask, &xev);
+}
+
+
+static void
+placeWin (CompWindow *window,
+     	  int        x,
+	  int        y,
+	  int        *new_x,
+	  int        *new_y)
 {
     CompWindow *wi;
     GList      *windows;
@@ -1244,9 +1157,9 @@ placeWindow (CompWindow *window,
 	}
     }
 
-    if ((window->type == CompWindowTypeDialogMask ||
-	 window->type == CompWindowTypeModalDialogMask) &&
-	window->transientFor != None)
+    if (window->transientFor &&
+	(window->type & (CompWindowTypeDialogMask |
+			 CompWindowTypeModalDialogMask)))
     {
 	/* Center horizontally, at top of parent vertically */
 
@@ -1278,24 +1191,37 @@ placeWindow (CompWindow *window,
 	    y += window->input.top;
 
 	    /* clip to screen if parent is visible in current viewport */
-	    if (parent->serverX < parent->screen->width &&
-		parent->serverX + parent->screen->width > 0)
+	    if (parent->serverX < parent->screen->width   &&
+		parent->serverX + parent->serverWidth > 0 &&
+		parent->serverY < parent->screen->height  &&
+		parent->serverY + parent->serverHeight > 0)
 	    {
-		XRectangle area;
+		XRectangle        area;
+		int               output;
+		CompWindowExtents extents;
 
-		get_workarea_of_current_output_device (window->screen, &area);
+		output = outputDeviceForWindow (parent);
+		getWorkareaForOutput (window->screen, output, &area);
 
-		if (x + window_width > area.x + area.width)
-		    x = area.x + area.width - window_width;
-		if (y + window_height > area.y + area.height)
-		    y = area.y + area.height - window_height;
-		if (x < area.x) x = area.x;
-		if (y < area.y) y = area.y;
+		extents.left   = x - window->input.left;
+		extents.top    = y - window->input.top;
+		extents.right  = x + window_width + window->input.right;
+		extents.bottom = y + window_height + window->input.bottom;
+
+		if (extents.left < area.x)
+		    x += area.x - extents.left;
+		else if (extents.right > area.x + area.width)
+		    x += area.x + area.width - extents.right;
+
+		if (extents.top < area.y)
+		    y += area.y - extents.top;
+		else if (extents.bottom > area.y + area.height)
+		    y += area.y + area.height - extents.bottom;
 	    }
 
 	    avoid_being_obscured_as_second_modal_dialog (window, &x, &y);
 
-	    goto done_no_x_constraints;
+	    goto done_no_constraints;
 	}
     }
 
@@ -1350,10 +1276,23 @@ placeWindow (CompWindow *window,
     x = x0;
     y = y0;
 
-    if (!placeMatchPosition (window, &x, &y))
+    if (placeMatchPosition (window, &x, &y))
     {
-	switch (ps->placeMode) {
-	case PlaceModeCascade:
+	int output;
+
+	output = outputDeviceForGeometry (window->screen, x, y,
+					  window_width, window_height,
+					  window->serverBorderWidth);
+
+	getWorkareaForOutput (window->screen, output, &work_area);
+
+	work_area.x += x0;
+	work_area.y += y0;
+    }
+    else
+    {
+	switch (ps->opt[PLACE_SCREEN_OPTION_MODE].value.i) {
+	case PLACE_MODE_CASCADE:
 	    if (find_first_fit (window, windows, x, y, &x, &y))
 		goto done_check_denied_focus;
 
@@ -1362,37 +1301,23 @@ placeWindow (CompWindow *window,
 	     */
 	    find_next_cascade (window, windows, x, y, &x, &y);
 	    break;
-	case PlaceModeCentered:
+	case PLACE_MODE_CENTERED:
 	    placeCentered (window, &work_area, &x, &y);
 	    break;
-	case PlaceModeRandom:
+	case PLACE_MODE_RANDOM:
 	    placeRandom (window, &work_area, &x, &y);
 	    break;
-	case PlaceModeSmart:
+	case PLACE_MODE_SMART:
 	    placeSmart (window, &work_area, &x, &y);
 	    break;
-	case PlaceModeMaximize:
+	case PLACE_MODE_MAXIMIZE:
 	    maximizeWindow (window, MAXIMIZE_STATE);
 	    break;
+	default:
+	    break;
 	}
     }
 
-    /* Maximize windows if they are too big for their work area (bit of
-     * a hack here). Assume undecorated windows probably don't intend to
-     * be maximized.
-     */
-    if ((window->actions & MAXIMIZE_STATE) == MAXIMIZE_STATE &&
-	(window->mwmDecor & (MwmDecorAll | MwmDecorTitle))   &&
-	!(window->state & CompWindowStateFullscreenMask))
-    {
-	XRectangle outer;
-
-	getOuterRectOfWindow (window, &outer);
-
-	if (outer.width >= work_area.width && outer.height >= work_area.height)
-	    maximizeWindow (window, MAXIMIZE_STATE);
-    }
-
 done_check_denied_focus:
     /* If the window is being denied focus and isn't a transient of the
      * focus window, we do NOT want it to overlap with the focus window
@@ -1411,8 +1336,8 @@ done_check_denied_focus:
 	{
 	    XRectangle wr, fwr, overlap;
 
-	    getOuterRectOfWindow (window, &wr);
-	    getOuterRectOfWindow (focus_window, &fwr);
+	    get_outer_rect_of_window (window, &wr);
+	    get_outer_rect_of_window (focus_window, &fwr);
 
 	    /* No need to do anything if the window doesn't overlap at all */
 	    found_fit = !rectangleIntersect (&wr, &fwr, &overlap);
@@ -1447,13 +1372,28 @@ done_check_denied_focus:
     g_list_free (windows);
 
 done:
+    /* Maximize windows if they are too big for their work area (bit of
+     * a hack here). Assume undecorated windows probably don't intend to
+     * be maximized.
+     */
+    if ((window->actions & MAXIMIZE_STATE) == MAXIMIZE_STATE &&
+	(window->mwmDecor & (MwmDecorAll | MwmDecorTitle))   &&
+	!(window->state & CompWindowStateFullscreenMask))
+    {
+	XRectangle outer;
+
+	get_outer_rect_of_window (window, &outer);
+
+	if (outer.width >= work_area.width && outer.height >= work_area.height)
+	    maximizeWindow (window, MAXIMIZE_STATE);
+    }
+
     if (x + window_width + window->input.right > work_area.x + work_area.width)
 	x = work_area.x + work_area.width - window_width - window->input.right;
 
     if (x - window->input.left < work_area.x)
 	x = work_area.x + window->input.left;
 
-done_no_x_constraints:
     if (y + window_height + window->input.bottom >
 	work_area.y + work_area.height)
 	y = work_area.y + work_area.height
@@ -1467,45 +1407,172 @@ done_no_constraints:
     *new_y = y;
 }
 
+static void
+placeValidateWindowResizeRequest (CompWindow     *w,
+				  unsigned int   *mask,
+				  XWindowChanges *xwc)
+{
+    Bool       checkPlacement = FALSE;
+    CompScreen *s = w->screen;
+
+    PLACE_SCREEN (s);
+
+    UNWRAP (ps, s, validateWindowResizeRequest);
+    (*s->validateWindowResizeRequest) (w, mask, xwc);
+    WRAP (ps, s, validateWindowResizeRequest,
+	  placeValidateWindowResizeRequest);
+
+    if (w->type & (CompWindowTypeSplashMask      |
+		   CompWindowTypeDialogMask      |
+		   CompWindowTypeModalDialogMask |
+		   CompWindowTypeNormalMask))
+    {
+	if (!(w->state & CompWindowStateFullscreenMask))
+	{
+	    if (!(w->sizeHints.flags & USPosition))
+		checkPlacement = TRUE;
+	}
+    }
+
+    if (checkPlacement)
+    {
+	XRectangle workArea;
+	int        x, y, left, right, top, bottom;
+	int        output;
+
+	/* left, right, top, bottom target coordinates, clamped to viewport
+	   sizes as we don't need to validate movements to other viewports;
+	   we are only interested in inner-viewport movements */
+	x = xwc->x % s->width;
+	if (x < 0)
+	    x += s->width;
+
+	y = xwc->y % s->height;
+	if (y < 0)
+	    y += s->height;
+
+	left   = x - w->input.left;
+	right  = x + xwc->width + w->input.right;
+	top    = y - w->input.top;
+	bottom = y + xwc->height + w->input.bottom;
+
+	output = outputDeviceForGeometry (s,
+					  xwc->x, xwc->y,
+					  xwc->width, xwc->height,
+					  w->serverBorderWidth);
+
+	getWorkareaForOutput (s, output, &workArea);
+
+	if (xwc->width >= workArea.width &&
+	    xwc->height >= workArea.height)
+	{
+	    placeSendWindowMaximizationRequest (w);
+	}
+
+	if ((right - left) > workArea.width)
+	{
+	    left  = workArea.x;
+	    right = left + workArea.width;
+	}
+	else
+	{
+	    if (left < workArea.x)
+	    {
+		right += workArea.x - left;
+		left  = workArea.x;
+	    }
+
+	    if (right > (workArea.x + workArea.width))
+	    {
+		left -= right - (workArea.x + workArea.width);
+		right = workArea.x + workArea.width;
+	    }
+	}
+
+	if ((bottom - top) > workArea.height)
+	{
+	    top    = workArea.y;
+	    bottom = top + workArea.height;
+	}
+	else
+	{
+	    if (top < workArea.y)
+	    {
+		bottom += workArea.y - top;
+		top    = workArea.y;
+	    }
+
+	    if (bottom > (workArea.y + workArea.height))
+	    {
+		top   -= bottom - (workArea.y + workArea.height);
+		bottom = workArea.y + workArea.height;
+	    }
+	}
+
+	/* bring left/right/top/bottom to actual window coordinates */
+	left   += w->input.left;
+	right  -= w->input.right;
+	top    += w->input.top;
+	bottom -= w->input.bottom;
+
+	if (left != x)
+	{
+	    xwc->x += left - x;
+	    *mask  |= CWX;
+	}
+
+	if (top != y)
+	{
+	    xwc->y += top - y;
+	    *mask  |= CWY;
+	}
+
+	if ((right - left) != xwc->width)
+	{
+	    xwc->width = right - left;
+	    *mask      |= CWWidth;
+	}
+
+	if ((bottom - top) != xwc->height)
+	{
+	    xwc->height = bottom - top;
+	    *mask       |= CWHeight;
+	}
+    }
+}
+
 static Bool
-placeDamageWindowRect (CompWindow *w,
-		       Bool	  initial,
-		       BoxPtr     rect)
+placePlaceWindow (CompWindow *w,
+		  int        x,
+		  int        y,
+		  int        *newX,
+		  int        *newY)
 {
     Bool status;
 
     PLACE_SCREEN (w->screen);
 
-    UNWRAP (ps, w->screen, damageWindowRect);
-    status = (*w->screen->damageWindowRect) (w, initial, rect);
-    WRAP (ps, w->screen, damageWindowRect, placeDamageWindowRect);
+    UNWRAP (ps, w->screen, placeWindow);
+    status = (*w->screen->placeWindow) (w, x, y, newX, newY);
+    WRAP (ps, w->screen, placeWindow, placePlaceWindow);
 
-    if (initial && !w->attrib.override_redirect && !w->placed)
+    if (!status)
     {
 	int viewportX, viewportY;
-	int newX, newY;
 
-	placeWindow (w, w->serverX, w->serverY, &newX, &newY);
+	placeWin (w, x, y, newX, newY);
 
 	if (placeMatchViewport (w, &viewportX, &viewportY))
 	{
 	    viewportX = MAX (MIN (viewportX, w->screen->hsize), 0);
 	    viewportY = MAX (MIN (viewportY, w->screen->vsize), 0);
 
-	    newX += (viewportX - w->screen->x) * w->screen->width;
-	    newY += (viewportY - w->screen->y) * w->screen->height;
-	}
-
-	w->placed = TRUE;
-
-	if (newX != w->serverX || newY != w->serverY)
-	{
-	    moveWindow (w, newX - w->attrib.x, newY - w->attrib.y, FALSE, TRUE);
-	    syncWindowPosition (w);
+	    *newX += (viewportX - w->screen->x) * w->screen->width;
+	    *newY += (viewportY - w->screen->y) * w->screen->height;
 	}
     }
 
-    return status;
+    return TRUE;
 }
 
 static Bool
@@ -1514,6 +1581,9 @@ placeInitDisplay (CompPlugin  *p,
 {
     PlaceDisplay *pd;
 
+    if (!checkPluginABI ("core", CORE_ABIVERSION))
+	return FALSE;
+
     pd = malloc (sizeof (PlaceDisplay));
     if (!pd)
 	return FALSE;
@@ -1525,7 +1595,7 @@ placeInitDisplay (CompPlugin  *p,
 	return FALSE;
     }
 
-    d->privates[displayPrivateIndex].ptr = pd;
+    d->base.privates[displayPrivateIndex].ptr = pd;
 
     return TRUE;
 }
@@ -1541,6 +1611,17 @@ placeFiniDisplay (CompPlugin  *p,
     free (pd);
 }
 
+static const CompMetadataOptionInfo placeScreenOptionInfo[] = {
+    { "workarounds", "bool", 0, 0, 0 },
+    { "mode", "int", RESTOSTRING (0, PLACE_MODE_LAST), 0, 0 },
+    { "position_matches", "list", "<type>match</type>", 0, 0 },
+    { "position_x_values", "list", "<type>int</type>", 0, 0 },
+    { "position_y_values", "list", "<type>int</type>", 0, 0 },
+    { "viewport_matches", "list", "<type>match</type>", 0, 0 },
+    { "viewport_x_values", "list", "<type>int</type>", 0, 0 },
+    { "viewport_y_values", "list", "<type>int</type>", 0, 0 }
+};
+
 static Bool
 placeInitScreen (CompPlugin *p,
 		 CompScreen *s)
@@ -1553,13 +1634,21 @@ placeInitScreen (CompPlugin *p,
     if (!ps)
 	return FALSE;
 
-    placeScreenInitOptions (ps);
+    if (!compInitScreenOptionsFromMetadata (s,
+					    &placeMetadata,
+					    placeScreenOptionInfo,
+					    ps->opt,
+					    PLACE_SCREEN_OPTION_NUM))
+    {
+	free (ps);
+	return FALSE;
+    }
 
-    WRAP (ps, s, damageWindowRect, placeDamageWindowRect);
+    WRAP (ps, s, placeWindow, placePlaceWindow);
+    WRAP (ps, s, validateWindowResizeRequest,
+	  placeValidateWindowResizeRequest);
 
-    s->privates[pd->screenPrivateIndex].ptr = ps;
-
-    placeUpdateMode (s);
+    s->base.privates[pd->screenPrivateIndex].ptr = ps;
 
     return TRUE;
 }
@@ -1570,17 +1659,88 @@ placeFiniScreen (CompPlugin *p,
 {
     PLACE_SCREEN (s);
 
-    UNWRAP (ps, s, damageWindowRect);
+    UNWRAP (ps, s, placeWindow);
+    UNWRAP (ps, s, validateWindowResizeRequest);
+
+    compFiniScreenOptions (s, ps->opt, PLACE_SCREEN_OPTION_NUM);
 
     free (ps);
 }
 
+static CompBool
+placeInitObject (CompPlugin *p,
+		 CompObject *o)
+{
+    static InitPluginObjectProc dispTab[] = {
+	(InitPluginObjectProc) 0, /* InitCore */
+	(InitPluginObjectProc) placeInitDisplay,
+	(InitPluginObjectProc) placeInitScreen
+    };
+
+    RETURN_DISPATCH (o, dispTab, ARRAY_SIZE (dispTab), TRUE, (p, o));
+}
+
+static void
+placeFiniObject (CompPlugin *p,
+		 CompObject *o)
+{
+    static FiniPluginObjectProc dispTab[] = {
+	(FiniPluginObjectProc) 0, /* FiniCore */
+	(FiniPluginObjectProc) placeFiniDisplay,
+	(FiniPluginObjectProc) placeFiniScreen
+    };
+
+    DISPATCH (o, dispTab, ARRAY_SIZE (dispTab), (p, o));
+}
+
+static CompOption *
+placeGetObjectOptions (CompPlugin *plugin,
+		       CompObject *object,
+		       int	  *count)
+{
+    static GetPluginObjectOptionsProc dispTab[] = {
+	(GetPluginObjectOptionsProc) 0, /* GetCoreOptions */
+	(GetPluginObjectOptionsProc) 0, /* GetDisplayOptions */
+	(GetPluginObjectOptionsProc) placeGetScreenOptions
+    };
+
+    RETURN_DISPATCH (object, dispTab, ARRAY_SIZE (dispTab),
+		     (void *) (*count = 0), (plugin, object, count));
+}
+
+static CompBool
+placeSetObjectOption (CompPlugin      *plugin,
+		      CompObject      *object,
+		      const char      *name,
+		      CompOptionValue *value)
+{
+    static SetPluginObjectOptionProc dispTab[] = {
+	(SetPluginObjectOptionProc) 0, /* SetCoreOption */
+	(SetPluginObjectOptionProc) 0, /* SetDisplayOption */
+	(SetPluginObjectOptionProc) placeSetScreenOption
+    };
+
+    RETURN_DISPATCH (object, dispTab, ARRAY_SIZE (dispTab), FALSE,
+		     (plugin, object, name, value));
+}
+
 static Bool
 placeInit (CompPlugin *p)
 {
+    if (!compInitPluginMetadataFromInfo (&placeMetadata,
+					 p->vTable->name, 0, 0,
+					 placeScreenOptionInfo,
+					 PLACE_SCREEN_OPTION_NUM))
+	return FALSE;
+
     displayPrivateIndex = allocateDisplayPrivateIndex ();
     if (displayPrivateIndex < 0)
+    {
+	compFiniMetadata (&placeMetadata);
 	return FALSE;
+    }
+
+    compAddMetadataFromFile (&placeMetadata, p->vTable->name);
 
     return TRUE;
 }
@@ -1588,42 +1748,29 @@ placeInit (CompPlugin *p)
 static void
 placeFini (CompPlugin *p)
 {
-    if (displayPrivateIndex >= 0)
-	freeDisplayPrivateIndex (displayPrivateIndex);
+    freeDisplayPrivateIndex (displayPrivateIndex);
+    compFiniMetadata (&placeMetadata);
 }
 
-static int
-placeGetVersion (CompPlugin *plugin,
-		 int	    version)
+static CompMetadata *
+placeGetMetadata (CompPlugin *plugin)
 {
-    return ABIVERSION;
+    return &placeMetadata;
 }
 
 static CompPluginVTable placeVTable = {
     "place",
-    N_("Place Windows"),
-    N_("Place windows at appropriate positions when mapped"),
-    placeGetVersion,
+    placeGetMetadata,
     placeInit,
     placeFini,
-    placeInitDisplay,
-    placeFiniDisplay,
-    placeInitScreen,
-    placeFiniScreen,
-    0, /* InitWindow */
-    0, /* FiniWindow */
-    0, /* GetDisplayOptions */
-    0, /* SetDisplayOption */
-    placeGetScreenOptions,
-    placeSetScreenOption,
-    0, /* Deps */
-    0, /* nDeps */
-    0, /* Features */
-    0  /* nFeatures */
+    placeInitObject,
+    placeFiniObject,
+    placeGetObjectOptions,
+    placeSetObjectOption
 };
 
 CompPluginVTable *
-getCompPluginInfo (void)
+getCompPluginInfo20070830 (void)
 {
     return &placeVTable;
 }
