@@ -32,7 +32,7 @@
 #include <X11/extensions/Xrandr.h>
 #include <X11/extensions/Xfixes.h>
 
-#include <compiz.h>
+#include <compiz-core.h>
 
 static Window xdndWindow = None;
 static Window edgeWindow = None;
@@ -74,9 +74,6 @@ handleWindowDamageRect (CompWindow *w,
 	damageScreenRegion (w->screen, &region);
     }
 
-    if (!w->attrib.override_redirect)
-	w->placed = TRUE;
-
     if (initial)
 	damageWindowOutputExtents (w);
 }
@@ -142,7 +139,7 @@ moveInputFocusToOtherWindow (CompWindow *w)
 		moveInputFocusToWindow (ancestor);
 	    }
 	    else
-		focusDefaultWindow (display);
+		focusDefaultWindow (w->screen);
 	}
 	else if (w->type & (CompWindowTypeDialogMask |
 			    CompWindowTypeModalDialogMask))
@@ -177,10 +174,10 @@ moveInputFocusToOtherWindow (CompWindow *w)
 		moveInputFocusToWindow (focus);
 	    }
 	    else
-		focusDefaultWindow (display);
+		focusDefaultWindow (w->screen);
 	}
 	else
-	    focusDefaultWindow (display);
+	    focusDefaultWindow (w->screen);
     }
 }
 
@@ -188,28 +185,28 @@ static Bool
 autoRaiseTimeout (void *closure)
 {
     CompDisplay *display = closure;
+    CompWindow  *w = findWindowAtDisplay (display, display->activeWindow);
 
-    if (display->autoRaiseWindow == display->activeWindow)
+    if (display->autoRaiseWindow == display->activeWindow ||
+	(w && (display->autoRaiseWindow == w->transientFor)))
     {
-	CompWindow *w;
-
 	w = findWindowAtDisplay (display, display->autoRaiseWindow);
 	if (w)
-	    updateWindowAttributes (w, FALSE);
+	    updateWindowAttributes (w, CompStackingUpdateModeNormal);
     }
 
     return FALSE;
 }
 
-#define REAL_MOD_MASK (ShiftMask | ControlMask | Mod1Mask | \
-		       Mod2Mask | Mod3Mask | Mod4Mask | Mod5Mask)
+#define REAL_MOD_MASK (ShiftMask | ControlMask | Mod1Mask | Mod2Mask | \
+		       Mod3Mask | Mod4Mask | Mod5Mask | CompNoMask)
 
 static Bool
 isCallBackBinding (CompOption	   *option,
 		   CompBindingType type,
 		   CompActionState state)
 {
-    if (option->type != CompOptionTypeAction)
+    if (!isActionOption (option))
 	return FALSE;
 
     if (!(option->value.action.type & type))
@@ -289,6 +286,7 @@ triggerButtonPressBindings (CompDisplay *d,
 	    if (edgeWindow == s->screenEdge[i].id)
 	    {
 		edge = 1 << i;
+		argument[1].value.i = d->activeWindow;
 		break;
 	    }
 	}
@@ -311,13 +309,18 @@ triggerButtonPressBindings (CompDisplay *d,
 
 	if (edge)
 	{
-	    if (isInitiateBinding (option, CompBindingTypeEdgeButton, state,
-				   &action))
+	    if (isInitiateBinding (option, CompBindingTypeEdgeButton,
+				   state | CompActionStateInitEdge, &action))
 	    {
-		if (action->edgeMask & edge)
+		if ((action->button.button == event->xbutton.button) &&
+		    (action->edgeMask & edge))
 		{
-		    if (action->edgeButton == event->xbutton.button)
-			if ((*action->initiate) (d, action,
+		    bindMods = virtualToRealModMask (d,
+						     action->button.modifiers);
+
+		    if ((bindMods & modMask) ==
+			(event->xbutton.state & modMask))
+			if ((*action->initiate) (d, action, state |
 						 CompActionStateInitEdge,
 						 argument, nArgument))
 			    return TRUE;
@@ -340,11 +343,12 @@ triggerButtonReleaseBindings (CompDisplay *d,
 			      int	  nArgument)
 {
     CompActionState state = CompActionStateTermButton;
+    CompBindingType type  = CompBindingTypeButton | CompBindingTypeEdgeButton;
     CompAction	    *action;
 
     while (nOption--)
     {
-	if (isTerminateBinding (option, CompBindingTypeButton, state, &action))
+	if (isTerminateBinding (option, type, state, &action))
 	{
 	    if (action->button.button == event->xbutton.button)
 	    {
@@ -385,7 +389,7 @@ triggerKeyPressBindings (CompDisplay *d,
 
 	while (n--)
 	{
-	    if (o->type == CompOptionTypeAction)
+	    if (isActionOption (o))
 	    {
 		if (o->value.action.terminate)
 		    (*o->value.action.terminate) (d, &o->value.action,
@@ -411,7 +415,7 @@ triggerKeyPressBindings (CompDisplay *d,
 		if ((bindMods & modMask) == (event->xkey.state & modMask))
 		    if ((*action->initiate) (d, action, state,
 					     argument, nArgument))
-			break;
+			return TRUE;
 	    }
 	    else if (!d->xkbEvent && action->key.keycode == 0)
 	    {
@@ -536,7 +540,8 @@ isBellAction (CompOption      *option,
 	      CompActionState state,
 	      CompAction      **action)
 {
-    if (option->type != CompOptionTypeAction)
+    if (option->type != CompOptionTypeAction &&
+	option->type != CompOptionTypeBell)
 	return FALSE;
 
     if (!option->value.action.bell)
@@ -582,7 +587,9 @@ isEdgeAction (CompOption      *option,
 	      CompActionState state,
 	      unsigned int    edge)
 {
-    if (option->type != CompOptionTypeAction)
+    if (option->type != CompOptionTypeAction &&
+	option->type != CompOptionTypeButton &&
+	option->type != CompOptionTypeEdge)
 	return FALSE;
 
     if (!(option->value.action.edgeMask & edge))
@@ -597,6 +604,7 @@ isEdgeAction (CompOption      *option,
 static Bool
 isEdgeEnterAction (CompOption      *option,
 		   CompActionState state,
+		   CompActionState delayState,
 		   unsigned int    edge,
 		   CompAction      **action)
 {
@@ -609,6 +617,18 @@ isEdgeEnterAction (CompOption      *option,
     if (!option->value.action.initiate)
 	return FALSE;
 
+    if (delayState)
+    {
+	if ((option->value.action.state & CompActionStateNoEdgeDelay) !=
+	    (delayState & CompActionStateNoEdgeDelay))
+	{
+	    /* ignore edge actions which shouldn't be delayed when invoking
+	       undelayed edges (or vice versa) */
+	    return FALSE;
+	}
+    }
+
+
     *action = &option->value.action;
 
     return TRUE;
@@ -636,6 +656,7 @@ triggerEdgeEnterBindings (CompDisplay	  *d,
 			  CompOption	  *option,
 			  int		  nOption,
 			  CompActionState state,
+			  CompActionState delayState,
 			  unsigned int	  edge,
 			  CompOption	  *argument,
 			  int		  nArgument)
@@ -644,7 +665,7 @@ triggerEdgeEnterBindings (CompDisplay	  *d,
 
     while (nOption--)
     {
-	if (isEdgeEnterAction (option, state, edge, &action))
+	if (isEdgeEnterAction (option, state, delayState, edge, &action))
 	{
 	    if ((*action->initiate) (d, action, state, argument, nArgument))
 		return TRUE;
@@ -681,152 +702,237 @@ triggerEdgeLeaveBindings (CompDisplay	  *d,
     return FALSE;
 }
 
+static Bool
+triggerAllEdgeEnterBindings (CompDisplay     *d,
+			     CompActionState state,
+			     CompActionState delayState,
+			     unsigned int    edge,
+			     CompOption	     *argument,
+			     int	     nArgument)
+{
+    CompOption *option;
+    int        nOption;
+    CompPlugin *p;
+
+    for (p = getPlugins (); p; p = p->next)
+    {
+	if (p->vTable->getObjectOptions)
+	{
+	    option = (*p->vTable->getObjectOptions) (p, &d->base, &nOption);
+	    if (triggerEdgeEnterBindings (d,
+					  option, nOption,
+					  state, delayState, edge,
+					  argument, nArgument))
+	    {
+		return TRUE;
+	    }
+	}
+    }
+    return FALSE;
+}
+
+static Bool
+delayedEdgeTimeout (void *closure)
+{
+    CompDelayedEdgeSettings *settings = (CompDelayedEdgeSettings *) closure;
+
+    triggerAllEdgeEnterBindings (settings->d,
+				 settings->state,
+				 ~CompActionStateNoEdgeDelay,
+				 settings->edge,
+				 settings->option,
+				 settings->nOption);
+
+    free (settings);
+
+    return FALSE;
+}
+
+static Bool
+triggerEdgeEnter (CompDisplay     *d,
+		  unsigned int    edge,
+		  CompActionState state,
+		  CompOption      *argument,
+		  unsigned int    nArgument)
+{
+    int                     delay;
+    CompDelayedEdgeSettings *delayedSettings = NULL;
+
+    delay = d->opt[COMP_DISPLAY_OPTION_EDGE_DELAY].value.i;
+
+    if (nArgument > 7)
+	nArgument = 7;
+
+    if (delay > 0)
+    {
+	delayedSettings = malloc (sizeof (CompDelayedEdgeSettings));
+	if (delayedSettings)
+	{
+	    delayedSettings->d       = d;
+	    delayedSettings->edge    = edge;
+	    delayedSettings->state   = state;
+	    delayedSettings->nOption = nArgument;
+	}
+    }
+
+    if (delayedSettings)
+    {
+	CompActionState delayState;
+	int             i;
+
+	for (i = 0; i < nArgument; i++)
+	    delayedSettings->option[i] = argument[i];
+
+	d->edgeDelayHandle = compAddTimeout (delay,
+					     delayedEdgeTimeout,
+					     delayedSettings);
+
+	delayState = CompActionStateNoEdgeDelay;
+	if (triggerAllEdgeEnterBindings (d, state, delayState,
+					 edge, argument, nArgument))
+	    return TRUE;
+    }
+    else
+    {
+	if (triggerAllEdgeEnterBindings (d, state, 0, edge,
+					 argument, nArgument))
+	    return TRUE;
+    }
+
+    return FALSE;
+}
+
 static Bool
 handleActionEvent (CompDisplay *d,
 		   XEvent      *event)
 {
+    CompObject *obj = &d->base;
     CompOption *option;
     int	       nOption;
     CompPlugin *p;
-    CompOption o[7];
+    CompOption o[8];
 
     o[0].type = CompOptionTypeInt;
-    o[0].name = "window";
+    o[0].name = "event_window";
 
     o[1].type = CompOptionTypeInt;
-    o[1].name = "modifiers";
+    o[1].name = "window";
 
     o[2].type = CompOptionTypeInt;
-    o[2].name = "x";
+    o[2].name = "modifiers";
 
     o[3].type = CompOptionTypeInt;
-    o[3].name = "y";
+    o[3].name = "x";
 
     o[4].type = CompOptionTypeInt;
-    o[4].name = "root";
+    o[4].name = "y";
+
+    o[5].type = CompOptionTypeInt;
+    o[5].name = "root";
 
     switch (event->type) {
     case ButtonPress:
 	o[0].value.i = event->xbutton.window;
-	o[1].value.i = event->xbutton.state;
-	o[2].value.i = event->xbutton.x_root;
-	o[3].value.i = event->xbutton.y_root;
-	o[4].value.i = event->xbutton.root;
-
-	o[5].type    = CompOptionTypeInt;
-	o[5].name    = "button";
-	o[5].value.i = event->xbutton.button;
+	o[1].value.i = event->xbutton.window;
+	o[2].value.i = event->xbutton.state;
+	o[3].value.i = event->xbutton.x_root;
+	o[4].value.i = event->xbutton.y_root;
+	o[5].value.i = event->xbutton.root;
 
 	o[6].type    = CompOptionTypeInt;
-	o[6].name    = "time";
-	o[6].value.i = event->xbutton.time;
+	o[6].name    = "button";
+	o[6].value.i = event->xbutton.button;
+
+	o[7].type    = CompOptionTypeInt;
+	o[7].name    = "time";
+	o[7].value.i = event->xbutton.time;
 
 	for (p = getPlugins (); p; p = p->next)
 	{
-	    if (p->vTable->getDisplayOptions)
-	    {
-		option = (*p->vTable->getDisplayOptions) (d, &nOption);
-		if (triggerButtonPressBindings (d, option, nOption, event,
-						o, 7))
-		    return TRUE;
-	    }
+	    if (!p->vTable->getObjectOptions)
+		continue;
+
+	    option = (*p->vTable->getObjectOptions) (p, obj, &nOption);
+	    if (triggerButtonPressBindings (d, option, nOption, event, o, 8))
+		return TRUE;
 	}
-
-	option = compGetDisplayOptions (d, &nOption);
-	if (triggerButtonPressBindings (d, option, nOption, event, o, 7))
-	    return TRUE;
-
 	break;
     case ButtonRelease:
 	o[0].value.i = event->xbutton.window;
-	o[1].value.i = event->xbutton.state;
-	o[2].value.i = event->xbutton.x_root;
-	o[3].value.i = event->xbutton.y_root;
-	o[4].value.i = event->xbutton.root;
-
-	o[5].type    = CompOptionTypeInt;
-	o[5].name    = "button";
-	o[5].value.i = event->xbutton.button;
+	o[1].value.i = event->xbutton.window;
+	o[2].value.i = event->xbutton.state;
+	o[3].value.i = event->xbutton.x_root;
+	o[4].value.i = event->xbutton.y_root;
+	o[5].value.i = event->xbutton.root;
 
 	o[6].type    = CompOptionTypeInt;
-	o[6].name    = "time";
-	o[6].value.i = event->xbutton.time;
+	o[6].name    = "button";
+	o[6].value.i = event->xbutton.button;
+
+	o[7].type    = CompOptionTypeInt;
+	o[7].name    = "time";
+	o[7].value.i = event->xbutton.time;
 
 	for (p = getPlugins (); p; p = p->next)
 	{
-	    if (p->vTable->getDisplayOptions)
-	    {
-		option = (*p->vTable->getDisplayOptions) (d, &nOption);
-		if (triggerButtonReleaseBindings (d, option, nOption, event,
-						  o, 7))
-		    return TRUE;
-	    }
+	    if (!p->vTable->getObjectOptions)
+		continue;
+
+	    option = (*p->vTable->getObjectOptions) (p, obj, &nOption);
+	    if (triggerButtonReleaseBindings (d, option, nOption, event, o, 8))
+		return TRUE;
 	}
-
-	option = compGetDisplayOptions (d, &nOption);
-	if (triggerButtonReleaseBindings (d, option, nOption, event, o, 7))
-	    return TRUE;
-
 	break;
     case KeyPress:
-	o[0].value.i = d->activeWindow;
-	o[1].value.i = event->xkey.state;
-	o[2].value.i = event->xkey.x_root;
-	o[3].value.i = event->xkey.y_root;
-	o[4].value.i = event->xkey.root;
-
-	o[5].type    = CompOptionTypeInt;
-	o[5].name    = "keycode";
-	o[5].value.i = event->xkey.keycode;
+	o[0].value.i = event->xkey.window;
+	o[1].value.i = d->activeWindow;
+	o[2].value.i = event->xkey.state;
+	o[3].value.i = event->xkey.x_root;
+	o[4].value.i = event->xkey.y_root;
+	o[5].value.i = event->xkey.root;
 
 	o[6].type    = CompOptionTypeInt;
-	o[6].name    = "time";
-	o[6].value.i = event->xkey.time;
+	o[6].name    = "keycode";
+	o[6].value.i = event->xkey.keycode;
+
+	o[7].type    = CompOptionTypeInt;
+	o[7].name    = "time";
+	o[7].value.i = event->xkey.time;
 
 	for (p = getPlugins (); p; p = p->next)
 	{
-	    if (p->vTable->getDisplayOptions)
-	    {
-		option = (*p->vTable->getDisplayOptions) (d, &nOption);
-		if (triggerKeyPressBindings (d, option, nOption, event, o, 7))
-		    return TRUE;
-	    }
+	    if (!p->vTable->getObjectOptions)
+		continue;
+
+	    option = (*p->vTable->getObjectOptions) (p, obj, &nOption);
+	    if (triggerKeyPressBindings (d, option, nOption, event, o, 8))
+		return TRUE;
 	}
-
-	option = compGetDisplayOptions (d, &nOption);
-	if (triggerKeyPressBindings (d, option, nOption, event, o, 7))
-	    return TRUE;
-
 	break;
     case KeyRelease:
-	o[0].value.i = d->activeWindow;
-	o[1].value.i = event->xkey.state;
-	o[2].value.i = event->xkey.x_root;
-	o[3].value.i = event->xkey.y_root;
-	o[4].value.i = event->xkey.root;
-
-	o[5].type    = CompOptionTypeInt;
-	o[5].name    = "keycode";
-	o[5].value.i = event->xkey.keycode;
+	o[0].value.i = event->xkey.window;
+	o[1].value.i = d->activeWindow;
+	o[2].value.i = event->xkey.state;
+	o[3].value.i = event->xkey.x_root;
+	o[4].value.i = event->xkey.y_root;
+	o[5].value.i = event->xkey.root;
 
 	o[6].type    = CompOptionTypeInt;
-	o[6].name    = "time";
-	o[6].value.i = event->xkey.time;
+	o[6].name    = "keycode";
+	o[6].value.i = event->xkey.keycode;
+
+	o[7].type    = CompOptionTypeInt;
+	o[7].name    = "time";
+	o[7].value.i = event->xkey.time;
 
 	for (p = getPlugins (); p; p = p->next)
 	{
-	    if (p->vTable->getDisplayOptions)
-	    {
-		option = (*p->vTable->getDisplayOptions) (d, &nOption);
-		if (triggerKeyReleaseBindings (d, option, nOption, event, o, 7))
-		    return TRUE;
-	    }
+	    if (!p->vTable->getObjectOptions)
+		continue;
+	    option = (*p->vTable->getObjectOptions) (p, obj, &nOption);
+	    if (triggerKeyReleaseBindings (d, option, nOption, event, o, 8))
+		return TRUE;
 	}
-
-	option = compGetDisplayOptions (d, &nOption);
-	if (triggerKeyReleaseBindings (d, option, nOption, event, o, 7))
-	    return TRUE;
-
 	break;
     case EnterNotify:
 	if (event->xcrossing.mode   != NotifyGrab   &&
@@ -841,6 +947,16 @@ handleActionEvent (CompDisplay *d,
 	    if (!s)
 		return FALSE;
 
+	    if (d->edgeDelayHandle)
+	    {
+		void *closure;
+
+		closure = compRemoveTimeout (d->edgeDelayHandle);
+		if (closure)
+		    free (closure);
+		d->edgeDelayHandle = 0;
+	    }
+
 	    if (edgeWindow && edgeWindow != event->xcrossing.window)
 	    {
 		state = CompActionStateTermEdge;
@@ -857,31 +973,27 @@ handleActionEvent (CompDisplay *d,
 
 		edgeWindow = None;
 
-		o[0].value.i = d->activeWindow;
-		o[1].value.i = event->xcrossing.state;
-		o[2].value.i = event->xcrossing.x_root;
-		o[3].value.i = event->xcrossing.y_root;
-		o[4].value.i = event->xcrossing.root;
+		o[0].value.i = event->xcrossing.window;
+		o[1].value.i = d->activeWindow;
+		o[2].value.i = event->xcrossing.state;
+		o[3].value.i = event->xcrossing.x_root;
+		o[4].value.i = event->xcrossing.y_root;
+		o[5].value.i = event->xcrossing.root;
 
-		o[5].type    = CompOptionTypeInt;
-		o[5].name    = "time";
-		o[5].value.i = event->xcrossing.time;
+		o[6].type    = CompOptionTypeInt;
+		o[6].name    = "time";
+		o[6].value.i = event->xcrossing.time;
 
 		for (p = getPlugins (); p; p = p->next)
 		{
-		    if (p->vTable->getDisplayOptions)
-		    {
-			option = (*p->vTable->getDisplayOptions) (d, &nOption);
-			if (triggerEdgeLeaveBindings (d, option, nOption, state,
-						      edge, o, 6))
-			    return TRUE;
-		    }
-		}
+		    if (!p->vTable->getObjectOptions)
+			continue;
 
-		option = compGetDisplayOptions (d, &nOption);
-		if (triggerEdgeLeaveBindings (d, option, nOption, state,
-					      edge, o, 6))
-		    return TRUE;
+		    option = (*p->vTable->getObjectOptions) (p, obj, &nOption);
+		    if (triggerEdgeLeaveBindings (d, option, nOption, state,
+						  edge, o, 7))
+			return TRUE;
+		}
 	    }
 
 	    edge = 0;
@@ -901,33 +1013,22 @@ handleActionEvent (CompDisplay *d,
 
 		edgeWindow = event->xcrossing.window;
 
-		o[0].value.i = d->activeWindow;
-		o[1].value.i = event->xcrossing.state;
-		o[2].value.i = event->xcrossing.x_root;
-		o[3].value.i = event->xcrossing.y_root;
-		o[4].value.i = event->xcrossing.root;
+		o[0].value.i = event->xcrossing.window;
+		o[1].value.i = d->activeWindow;
+		o[2].value.i = event->xcrossing.state;
+		o[3].value.i = event->xcrossing.x_root;
+		o[4].value.i = event->xcrossing.y_root;
+		o[5].value.i = event->xcrossing.root;
 
-		o[5].type    = CompOptionTypeInt;
-		o[5].name    = "time";
-		o[5].value.i = event->xcrossing.time;
+		o[6].type    = CompOptionTypeInt;
+		o[6].name    = "time";
+		o[6].value.i = event->xcrossing.time;
 
-		for (p = getPlugins (); p; p = p->next)
-		{
-		    if (p->vTable->getDisplayOptions)
-		    {
-			option = (*p->vTable->getDisplayOptions) (d, &nOption);
-			if (triggerEdgeEnterBindings (d, option, nOption, state,
-						      edge, o, 6))
-			    return TRUE;
-		    }
-		}
-
-		option = compGetDisplayOptions (d, &nOption);
-		if (triggerEdgeEnterBindings (d, option, nOption, state,
-					      edge, o, 6))
+		if (triggerEdgeEnter (d, edge, state, o, 7))
 		    return TRUE;
 	    }
-	} break;
+	}
+	break;
     case ClientMessage:
 	if (event->xclient.message_type == d->xdndEnterAtom)
 	{
@@ -965,27 +1066,23 @@ handleActionEvent (CompDisplay *d,
 	    {
 		state = CompActionStateTermEdgeDnd;
 
-		o[0].value.i = d->activeWindow;
-		o[1].value.i = 0; /* fixme */
+		o[0].value.i = event->xclient.window;
+		o[1].value.i = d->activeWindow;
 		o[2].value.i = 0; /* fixme */
 		o[3].value.i = 0; /* fixme */
-		o[4].value.i = root;
+		o[4].value.i = 0; /* fixme */
+		o[5].value.i = root;
 
 		for (p = getPlugins (); p; p = p->next)
 		{
-		    if (p->vTable->getDisplayOptions)
-		    {
-			option = (*p->vTable->getDisplayOptions) (d, &nOption);
-			if (triggerEdgeLeaveBindings (d, option, nOption, state,
-						      edge, o, 5))
-			    return TRUE;
-		    }
-		}
+		    if (!p->vTable->getObjectOptions)
+			continue;
 
-		option = compGetDisplayOptions (d, &nOption);
-		if (triggerEdgeLeaveBindings (d, option, nOption, state,
-					      edge, o, 5))
-		    return TRUE;
+		    option = (*p->vTable->getObjectOptions) (p, obj, &nOption);
+		    if (triggerEdgeLeaveBindings (d, option, nOption, state,
+						  edge, o, 6))
+			return TRUE;
+		}
 	    }
 	}
 	else if (event->xclient.message_type == d->xdndPositionAtom)
@@ -1020,26 +1117,14 @@ handleActionEvent (CompDisplay *d,
 	    {
 		state = CompActionStateInitEdgeDnd;
 
-		o[0].value.i = d->activeWindow;
-		o[1].value.i = 0; /* fixme */
-		o[2].value.i = event->xclient.data.l[2] >> 16;
-		o[3].value.i = event->xclient.data.l[2] & 0xffff;
-		o[4].value.i = root;
+		o[0].value.i = event->xclient.window;
+		o[1].value.i = d->activeWindow;
+		o[2].value.i = 0; /* fixme */
+		o[3].value.i = event->xclient.data.l[2] >> 16;
+		o[4].value.i = event->xclient.data.l[2] & 0xffff;
+		o[5].value.i = root;
 
-		for (p = getPlugins (); p; p = p->next)
-		{
-		    if (p->vTable->getDisplayOptions)
-		    {
-			option = (*p->vTable->getDisplayOptions) (d, &nOption);
-			if (triggerEdgeEnterBindings (d, option, nOption, state,
-						      edge, o, 5))
-			    return TRUE;
-		    }
-		}
-
-		option = compGetDisplayOptions (d, &nOption);
-		if (triggerEdgeEnterBindings (d, option, nOption, state,
-					      edge, o, 5))
+		if (triggerEdgeEnter (d, edge, state, o, 6))
 		    return TRUE;
 	    }
 
@@ -1066,10 +1151,29 @@ handleActionEvent (CompDisplay *d,
 	    {
 		XkbStateNotifyEvent *stateEvent = (XkbStateNotifyEvent *) event;
 
-		option = compGetDisplayOptions (d, &nOption);
-
 		o[0].value.i = d->activeWindow;
-		o[1].value.i = stateEvent->mods;
+		o[1].value.i = d->activeWindow;
+		o[2].value.i = stateEvent->mods;
+
+		o[3].type    = CompOptionTypeInt;
+		o[3].name    = "time";
+		o[3].value.i = xkbEvent->time;
+
+		for (p = getPlugins (); p; p = p->next)
+		{
+		    if (!p->vTable->getObjectOptions)
+			continue;
+
+		    option = (*p->vTable->getObjectOptions) (p, obj, &nOption);
+		    if (triggerStateNotifyBindings (d, option, nOption,
+						    stateEvent, o, 4))
+			return TRUE;
+		}
+	    }
+	    else if (xkbEvent->xkb_type == XkbBellNotify)
+	    {
+		o[0].value.i = d->activeWindow;
+		o[1].value.i = d->activeWindow;
 
 		o[2].type    = CompOptionTypeInt;
 		o[2].name    = "time";
@@ -1077,44 +1181,13 @@ handleActionEvent (CompDisplay *d,
 
 		for (p = getPlugins (); p; p = p->next)
 		{
-		    if (p->vTable->getDisplayOptions)
-		    {
-			option = (*p->vTable->getDisplayOptions) (d, &nOption);
-			if (triggerStateNotifyBindings (d, option, nOption,
-							stateEvent, o, 3))
-			    return TRUE;
-		    }
+		    if (!p->vTable->getObjectOptions)
+			continue;
+
+		    option = (*p->vTable->getObjectOptions) (p, obj, &nOption);
+		    if (triggerBellNotifyBindings (d, option, nOption, o, 3))
+			return TRUE;
 		}
-
-		option = compGetDisplayOptions (d, &nOption);
-		if (triggerStateNotifyBindings (d, option, nOption, stateEvent,
-						o, 3))
-		    return TRUE;
-	    }
-	    else if (xkbEvent->xkb_type == XkbBellNotify)
-	    {
-		option = compGetDisplayOptions (d, &nOption);
-
-		o[0].value.i = d->activeWindow;
-
-		o[1].type    = CompOptionTypeInt;
-		o[1].name    = "time";
-		o[1].value.i = xkbEvent->time;
-
-		for (p = getPlugins (); p; p = p->next)
-		{
-		    if (p->vTable->getDisplayOptions)
-		    {
-			option = (*p->vTable->getDisplayOptions) (d, &nOption);
-			if (triggerBellNotifyBindings (d, option, nOption,
-						       o, 2))
-			    return TRUE;
-		    }
-		}
-
-		option = compGetDisplayOptions (d, &nOption);
-		if (triggerBellNotifyBindings (d, option, nOption, o, 2))
-		    return TRUE;
 	    }
 	}
 	break;
@@ -1125,8 +1198,8 @@ handleActionEvent (CompDisplay *d,
 
 void
 handleCompizEvent (CompDisplay *d,
-		   char        *pluginName,
-		   char        *eventName,
+		   const char  *pluginName,
+		   const char  *eventName,
 		   CompOption  *option,
 		   int         nOption)
 {
@@ -1172,7 +1245,10 @@ handleEvent (CompDisplay *d,
 
     switch (event->type) {
     case Expose:
-	s = findScreenAtDisplay (d, event->xexpose.window);
+	for (s = d->screens; s; s = s->next)
+	    if (s->output == event->xexpose.window)
+		break;
+
 	if (s)
 	{
 	    int more = event->xexpose.count + 1;
@@ -1264,6 +1340,9 @@ handleEvent (CompDisplay *d,
 	w = findWindowAtDisplay (d, event->xmap.window);
 	if (w)
 	{
+	    if (!w->attrib.override_redirect)
+		w->managed = TRUE;
+
 	    /* been shaded */
 	    if (w->height == 0)
 	    {
@@ -1290,9 +1369,9 @@ handleEvent (CompDisplay *d,
 		if (w->state & CompWindowStateHiddenMask)
 		{
 		    w->minimized = FALSE;
-		    w->state &= ~CompWindowStateHiddenMask;
 
-		    changeWindowState (w, w->state);
+		    changeWindowState (w,
+				       w->state & ~CompWindowStateHiddenMask);
 
 		    updateClientListForScreen (w->screen);
 		}
@@ -1348,7 +1427,8 @@ handleEvent (CompDisplay *d,
 		if (w)
 		{
 		    if (d->opt[COMP_DISPLAY_OPTION_RAISE_ON_CLICK].value.b)
-			updateWindowAttributes (w, TRUE);
+			updateWindowAttributes (w,
+					CompStackingUpdateModeAboveFullscreen);
 
 		    if (!(w->type & CompWindowTypeDockMask))
 			moveInputFocusToWindow (w);
@@ -1440,7 +1520,10 @@ handleEvent (CompDisplay *d,
 	{
 	    w = findWindowAtDisplay (d, event->xproperty.window);
 	    if (w)
+	    {
 		updateTransientHint (w);
+		recalcWindowActions (w);
+	    }
 	}
 	else if (event->xproperty.atom == d->wmClientLeaderAtom)
 	{
@@ -1575,7 +1658,15 @@ handleEvent (CompDisplay *d,
 	{
 	    w = findWindowAtDisplay (d, event->xclient.window);
 	    if (w)
-		activateWindow (w);
+	    {
+		/* use focus stealing prevention if request came from an
+		   application (which means data.l[0] is 1 */
+		if (event->xclient.data.l[0] != 1 ||
+		    allowWindowFocus (w, 0, event->xclient.data.l[1]))
+		{
+		    (*w->screen->activateWindow) (w);
+		}
+	    }
 	}
 	else if (event->xclient.message_type == d->winOpacityAtom)
 	{
@@ -1645,14 +1736,23 @@ handleEvent (CompDisplay *d,
 
 		if (wState != w->state)
 		{
-		    w->state = wState;
+		    CompStackingUpdateMode stackingUpdateMode;
+		    unsigned long          dState = wState ^ w->state;
 
-		    recalcWindowType (w);
-		    recalcWindowActions (w);
+		    stackingUpdateMode = CompStackingUpdateModeNone;
 
-		    changeWindowState (w, w->state);
+		    /* raise the window whenever its fullscreen state,
+		       above/below state or maximization state changed */
+		    if (dState & (CompWindowStateFullscreenMask |
+				  CompWindowStateAboveMask |
+				  CompWindowStateBelowMask |
+				  CompWindowStateMaximizedHorzMask |
+				  CompWindowStateMaximizedVertMask))
+			stackingUpdateMode = CompStackingUpdateModeNormal;
 
-		    updateWindowAttributes (w, FALSE);
+		    changeWindowState (w, wState);
+
+		    updateWindowAttributes (w, stackingUpdateMode);
 		}
 	    }
 	}
@@ -1703,11 +1803,11 @@ handleEvent (CompDisplay *d,
 
 		value.i = event->xclient.data.l[0] / s->width;
 
-		(*s->setScreenOption) (s, "hsize", &value);
+		(*core.setOptionForPlugin) (&s->base, "core", "hsize", &value);
 
 		value.i = event->xclient.data.l[1] / s->height;
 
-		(*s->setScreenOption) (s, "vsize", &value);
+		(*core.setOptionForPlugin) (&s->base, "core", "vsize", &value);
 	    }
 	}
 	else if (event->xclient.message_type == d->moveResizeWindowAtom)
@@ -1781,7 +1881,7 @@ handleEvent (CompDisplay *d,
 	else if (event->xclient.message_type == d->wmChangeStateAtom)
 	{
 	    w = findWindowAtDisplay (d, event->xclient.window);
-	    if (w && w->type & CompWindowTypeNormalMask)
+	    if (w)
 	    {
 		if (event->xclient.data.l[0] == IconicState)
 		    minimizeWindow (w);
@@ -1812,7 +1912,9 @@ handleEvent (CompDisplay *d,
 
 		value.i = event->xclient.data.l[0];
 
-		(*s->setScreenOption) (s, "number_of_desktops", &value);
+		(*core.setOptionForPlugin) (&s->base,
+					    "core", "number_of_desktops",
+					    &value);
 	    }
 	}
 	else if (event->xclient.message_type == d->currentDesktopAtom)
@@ -1835,15 +1937,34 @@ handleEvent (CompDisplay *d,
 	w = findWindowAtDisplay (d, event->xmaprequest.window);
 	if (w)
 	{
-	    if (w->minimized)
-		unminimizeWindow (w);
+	    XWindowAttributes attr;
+	    Bool              doMapProcessing = TRUE;
 
-	    (*w->screen->leaveShowDesktopMode) (w->screen, w);
+	    /* We should check the override_redirect flag here, because the
+	       client might have changed it while being unmapped. */
+	    if (XGetWindowAttributes (d->display, w->id, &attr))
+	    {
+		if (w->attrib.override_redirect != attr.override_redirect)
+		{
+		    w->attrib.override_redirect = attr.override_redirect;
+		    recalcWindowType (w);
+		    recalcWindowActions (w);
+
+		    (*d->matchPropertyChanged) (d, w);
+		}
+	    }
 
 	    w->managed = TRUE;
 
-	    if (!(w->state & CompWindowStateHiddenMask))
+	    if (w->state & CompWindowStateHiddenMask)
+		if (!w->minimized && !w->inShowDesktopMode)
+		    doMapProcessing = FALSE;
+
+	    if (doMapProcessing)
 	    {
+		Bool                   allowFocus;
+		CompStackingUpdateMode stackingMode;
+
 		w->initialViewportX = w->screen->x;
 		w->initialViewportY = w->screen->y;
 
@@ -1851,28 +1972,45 @@ handleEvent (CompDisplay *d,
 
 		applyStartupProperties (w->screen, w);
 
-		w->pendingMaps++;
-
-		XMapWindow (d->display, event->xmaprequest.window);
-
-		updateWindowAttributes (w, FALSE);
-
-		if (focusWindowOnMap (w))
+		if (!w->placed)
 		{
+		    int newX, newY;
+
+		    if ((*w->screen->placeWindow) (w, w->serverX, w->serverY,
+						   &newX, &newY))
+		    {
+			moveWindow (w,
+				    newX - w->attrib.x,
+				    newY - w->attrib.y,
+				    FALSE, TRUE);
+			syncWindowPosition (w);
+		    }
+
+		    w->placed   = TRUE;
+		}
+
+		allowFocus = allowWindowFocus (w, NO_FOCUS_MASK, 0);
+
+		if (!allowFocus && (w->type & ~NO_FOCUS_MASK))
+		    stackingMode = CompStackingUpdateModeInitialMapDeniedFocus;
+		else
+		    stackingMode = CompStackingUpdateModeInitialMap;
+
+		updateWindowAttributes (w, stackingMode);
+
+		if (w->minimized)
+		    unminimizeWindow (w);
+
+		(*w->screen->leaveShowDesktopMode) (w->screen, w);
+
+		if (!(w->state & CompWindowStateHiddenMask))
+		{
+		    w->pendingMaps++;
+		    XMapWindow (d->display, w->id);
+		}
+
+		if (allowFocus)
 		    moveInputFocusToWindow (w);
-		}
-		else if (w->type & ~CompWindowTypeSplashMask)
-		{
-		    CompWindow *p;
-
-		    for (p = w->prev; p; p = p->prev)
-			if (p->id == d->activeWindow)
-			    break;
-
-		    /* window is above active window so we should lower it */
-		    if (p)
-			restackWindowBelow (w, p);
-		}
 	    }
 
 	    setWindowProp (d, w->id, d->winDesktopAtom, w->desktop);
@@ -1897,6 +2035,46 @@ handleEvent (CompDisplay *d,
 	    xwc.border_width = event->xconfigurerequest.border_width;
 
 	    moveResizeWindow (w, &xwc, event->xconfigurerequest.value_mask, 0);
+
+	    if (event->xconfigurerequest.value_mask & CWStackMode)
+	    {
+		Window     above    = None;
+		CompWindow *sibling = NULL;
+
+		if (event->xconfigurerequest.value_mask & CWSibling)
+		{
+		    above   = event->xconfigurerequest.above;
+		    sibling = findTopLevelWindowAtDisplay (d, above);
+		}
+
+		switch (event->xconfigurerequest.detail) {
+		case Above:
+		    if (allowWindowFocus (w, NO_FOCUS_MASK, 0))
+		    {
+			if (above)
+			{
+			    if (sibling)
+				restackWindowAbove (w, sibling);
+			}
+			else
+			    raiseWindow (w);
+		    }
+		    break;
+		case Below:
+		    if (above)
+		    {
+			if (sibling)
+			    restackWindowBelow (w, sibling);
+		    }
+		    else
+			lowerWindow (w);
+		    break;
+		default:
+		    /* no handling of the TopIf, BottomIf, Opposite cases -
+		       there will hardly be any client using that */
+		    break;
+		}
+	    }
 	}
 	else
 	{
@@ -1943,9 +2121,7 @@ handleEvent (CompDisplay *d,
 		}
 
 		state &= ~CompWindowStateDemandsAttentionMask;
-
-		if (w->state != state)
-		    changeWindowState (w, state);
+		changeWindowState (w, state);
 	    }
 	}
 	break;
@@ -1996,7 +2172,12 @@ handleEvent (CompDisplay *d,
 				    compAddTimeout (delay, autoRaiseTimeout, d);
 			    }
 			    else
-				updateWindowAttributes (w, FALSE);
+			    {
+				CompStackingUpdateMode mode =
+				    CompStackingUpdateModeNormal;
+
+				updateWindowAttributes (w, mode);
+			    }
 			}
 		    }
 		}
@@ -2078,7 +2259,8 @@ handleEvent (CompDisplay *d,
 		}
 	    }
 	}
-	else if (event->type == d->randrEvent + RRScreenChangeNotify)
+	else if (d->randrExtension &&
+		 event->type == d->randrEvent + RRScreenChangeNotify)
 	{
 	    XRRScreenChangeNotifyEvent *rre;
 
